import asyncio
import functools

from gcloud.aio.core.utils.pools import pools


def maybe_async(callable_, *args, **kwargs):

    """
    Turn a callable into a coroutine if it isn't
    """

    if asyncio.iscoroutine(callable_):
        return callable_

    return asyncio.coroutine(callable_)(*args, **kwargs)


def fire(callable_, *args, **kwargs):

    """
    Start a callable as a coroutine, and return it's future. The cool thing
    about this function is that (via maybe_async) it lets you treat synchronous
    and asynchronous callables the same (both as async), which simplifies code.
    """

    return asyncio.ensure_future(maybe_async(callable_, *args, **kwargs))


def auto(fn):

    """
    Decorate a function or method with this, and it will become a callable
    that can be scheduled in the event loop just by calling it. Normally you'd
    have to do an `asyncio.ensure_future(my_callable())`. Not you can just do
    `my_callable()`. Twisted has always let you do this, and now you can let
    asyncio do it as well (with a decorator, albeit...)
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):

        return fire(fn, *args, **kwargs)

    return wrapper


async def _call_later(delay, callable_, *args, **kwargs):

    """
    The bus stop, where we wait.
    """

    await asyncio.sleep(delay)

    fire(callable_, *args, **kwargs)


def call_later(delay, callable_, *args, **kwargs):

    """
    After :delay seconds, call :callable with :args and :kwargs; :callable can
    be a synchronous or asynchronous callable (a coroutine). Note that _this_
    function is synchronous - mission accomplished - it can be used from within
    any synchronous or asynchronous callable.
    """

    return fire(_call_later, delay, callable_, *args, **kwargs)


def complete(value=None):

    """
    An awaitable that's already completed.
    """

    future = asyncio.Future()
    future.set_result(value)

    return future


def in_thread(fn, loop=None):

    """
    Decorate a function to make it always run in a thread.
    """

    if pools.get('None') is None:
        raise Exception('Threading has been disabled for this service.')

    loop = loop or asyncio.get_event_loop()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):

        if kwargs:
            _fn = functools.partial(fn, **kwargs)
        else:
            _fn = fn

        return loop.run_in_executor(pools.get('thread'), fn, *args)

    return wrapper


def in_subprocess(fn, loop=None):

    """
    Decorate a function to make it always run in a subprocess.
    """

    loop = loop or asyncio.get_event_loop()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):

        if kwargs:
            _fn = functools.partial(fn, **kwargs)
        else:
            _fn = fn

        return pools.get('process').submit(fn, *args)

    return wrapper


async def noop(*args, **kwargs):  # pylint: disable=unused-argument

    await asyncio.sleep(0)



import asyncio
import sys
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.process import BrokenProcessPool


MULTIPROCESSING_POOL_SIZE = 1
# for pickling deeply nested objects, like data sci models, you
# may need to increase this from the default of 1000
RECURSION_LIMIT = 1000
THREADING_POOL_SIZE = 10


sys.setrecursionlimit(RECURSION_LIMIT)


class ProcessPool(object):

    def __init__(self, loop, *args, **kwargs):

        self.loop = loop or asyncio.get_event_loop()
        self.args = args
        self.kwargs = kwargs

        self.pool = None
        self.create_pool()

    def __getattr__(self, name):

        return getattr(self.pool, name)

    async def submit(self, fn, *args):

        pool_id = id(self.pool)

        try:

            result = await self.loop.run_in_executor(self.pool, fn, *args)
            return result

        except BrokenProcessPool:

            print('Pool broke.')
            self.create_pool(pool_id)

            # maybe do a call_later or call_soon here
            result = await self.submit(fn, *args)
            return result

    def create_pool(self, pool_id=None):

        if self.pool is None:
            print('Creating pool')
            self.pool = ProcessPoolExecutor(*self.args, **self.kwargs)
            return

        if pool_id == id(self.pool):
            print('Recreating pool.')
            self.pool = ProcessPoolExecutor(*self.args, **self.kwargs)
            return

        print('No need to recreate.')


def make_thread_pool():

    return ThreadPoolExecutor(
        max_workers=THREADING_POOL_SIZE
    )


def make_process_pool(loop=None):

    return ProcessPool(
        loop,
        max_workers=MULTIPROCESSING_POOL_SIZE
    )


pools = {
    'thread': make_thread_pool(),
    'process': make_process_pool()
}



import json


def json_read(file_name: str):

    with open(file_name, 'r') as f:
        data = f.read()

    return json.loads(data)


def extract_json_fields(content, spec):

    if 'error' in content:
        raise Exception('{}'.format(content))

    return {field: cast(content[field]) for field, cast in spec}



import base64


def clean_b64decode(payload):

    """
    https://en.wikipedia.org/wiki/Base64#URL_applications modified Base64
    for URL variants exist, where the + and / characters of standard
    Base64 are respectively replaced by - and _
    """

    return base64.b64decode(payload.replace('-', '+').replace('_', '/'))


def clean_b64encode(payload):

    """
    https://en.wikipedia.org/wiki/Base64#URL_applications modified Base64
    for URL variants exist, where the + and / characters of standard
    Base64 are respectively replaced by - and _
    """

    if not isinstance(payload, bytes):
        payload = payload.encode('utf-8')

    return (
        base64.b64encode(payload)
        .replace(b'+', b'-')
        .replace(b'/', b'_')
        .decode('utf-8')
    )



import json
from urllib.parse import quote_plus
from urllib.parse import urlencode

import aiohttp
from asyncio_extras.contextmanager import async_contextmanager


class HttpError(Exception):
    pass


@async_contextmanager
async def ensure_session(session):

    if session:
        yield session
    else:
        async with aiohttp.ClientSession() as session:
            yield session


async def delete(url, headers=None, params=None, timeout=60, session=None):

    async with ensure_session(session) as s:  # pylint: disable=not-async-context-manager

        response = await s.delete(
            url,
            headers=headers,
            params=params,
            timeout=timeout
        )

        phrase = await response.text()

    return response.status, phrase


async def post(url, payload=None, timeout=60, urlencoded=False,
               json_response=True, session=None, headers=None, params=None):
    # pylint: disable=too-many-arguments

    headers = headers or {}

    if urlencoded:

        if payload:
            payload = urlencode(payload, quote_via=quote_plus)

        headers['content-type'] = 'application/x-www-form-urlencoded'

    else:

        if payload:
            payload = json.dumps(payload)
            payload = payload.encode('utf-8')
            content_length = str(len(payload))
        else:
            content_length = '0'

        headers.update({
            'content-length': content_length,
            'content-type': 'application/json'
        })

    async with ensure_session(session) as s:  # pylint: disable=not-async-context-manager

        response = await s.post(
            url,
            data=payload,
            headers=headers,
            params=params,
            timeout=timeout
        )

        if json_response:
            content = await response.json()
        else:
            content = await response.text()

    return response.status, content


async def get(url, timeout=60, json_response=True, session=None, headers=None,
              params=None):
    # pylint: disable=too-many-arguments

    async with ensure_session(session) as s:  # pylint: disable=not-async-context-manager

        response = await s.get(
            url,
            headers=headers,
            params=params,
            timeout=timeout
        )

        if json_response:
            content = await response.json()
        else:
            content = await response.text()

    return response.status, content


async def put(*args, **kwargs):  # pylint: disable=unused-argument

    raise Exception('Not implemented.')


async def patch(url, payload=None, timeout=60, session=None, headers=None,
                params=None):
    # pylint: disable=too-many-arguments

    headers = headers or {}

    if payload:
        payload = json.dumps(payload)
        payload = payload.encode('utf-8')
        content_length = str(len(payload))
    else:
        content_length = '0'

    headers.update({
        'content-length': content_length,
        'content-type': 'application/json'
    })

    async with ensure_session(session) as s:  # pylint: disable=not-async-context-manager

        response = await s.patch(
            url,
            data=payload,
            headers=headers,
            params=params,
            timeout=timeout
        )

        phrase = await response.text()

    return response.status, phrase



import random
import time
import traceback


def sleep_retry(fn, args=None, kwargs=None, tries=2, sleep_duration=1.0):

    args = args or tuple()
    kwargs = kwargs or {}

    attempts = 0

    while attempts < tries:

        try:

            return fn(*args, **kwargs)

        except Exception:  # pylint: disable=broad-except

            attempts += 1

            print('After {} try(ies) could not create service: {}'.format(
                attempts, traceback.format_exc() or '?'
            ))

            time.sleep(sleep_duration)

    return False


def backoff(base=2, factor=1, max_value=None):

    """Generator for exponential decay.

    The Google docs warn to back off from polling their API if there is no
    work available in a task queue. So we does.

    # modified from:
    # https://github.com/litl/backoff/blob/master/backoff.py

        base: the mathematical base of the exponentiation operation
        factor: factor to multiply the exponentation by.
        max_value: The maximum value to yield. Once the value in the
             true exponential sequence exceeds this, the value
             of max_value will forever after be yielded.
    """

    n = 0

    # initial backoff delay is nothing
    yield 0

    while True:

        a = factor * base ** n

        if max_value is None or a < max_value:
            yield a
            n += 1
        else:
            yield max_value - random.random() * max_value / 10


SECRETS = []


def secret(o):

    SECRETS.append(o)
    return o


class Config(object):
    # pylint: disable=too-few-public-methods

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument

        self.__dict__.update(**kwargs)

    def __str__(self):

        lines = self.show(depth=0)

        return '\n'.join(lines)

    def show(self, depth=0):

        lines = []

        for k in sorted(self.__dict__):

            v = self.__dict__[k]

            if isinstance(v, self.__class__):
                lines.append('{}{}'.format(
                    '  ' * depth,
                    k
                ))
                lines.extend(v.show(depth=depth + 1))
                continue

            lines.append('{}{}: {}'.format(
                '  ' * depth,
                k,
                '<SECRET>' if v in SECRETS else v
            ))

        return lines
